/**
 * My Nixie Clock IN-12
 * Vladimir N. Shilov <shilow@ukr.net>
 * 2019.01.18
 */

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h >

#include "main.h"
#include "i2c.h"
#include "ds3231.h"
#include "rtos.h"
#include "event-system.h"

/* Defines */
#define TIMER2_HZ          250
#define TIMER2_PRESCALER  256
#define TIMER2_CS         (1<<CS22 | 1<<CS21 | 0<<CS20)
#define  TIMER2_CNT        (0x100 - (F_CPU / TIMER2_PRESCALER / TIMER2_HZ))

/* Variables */
static volatile uint8_t Digit[4] = {1, 2, 3, 4};
static rtc_t RTC;
volatile struct {
  uint8_t RTC_Int:  1;
  uint8_t rezerv:   7;
} Flag;
static btn_t Button[BTN_NUM] = {
  {0, evBTN1Pressed, evBTN1Holded, BUTTON1_PIN},
  {0, evBTN2Pressed, evBTN2Holded, BUTTON2_PIN},
  {0, evBTN3Pressed, evBTN3Holded, BUTTON3_PIN}
};

/* Constants */
/* Function prototypes */
static void Board_Init(void);
static void dotOn(void);
static void dotOff(void);
static void btnProcess(void);

void main(void) {
  /**
   * Локальные переменные
   */
  uint8_t event = 0;
  Flag.RTC_Int = 0;

  /**
   * Инициализация, настройка...
   */
  Board_Init();

  // Initialize Scheduler
  RTOS_Init();

  // Initialize I2C Bus and RTC
  I2C_Init();
  RTC_Init();
  RTC_ReadAll(&RTC);

  // Initialize Event State Machine
  ES_Init(stShowTime);
  showTime();

  RTOS_SetTask(btnProcess, 10, 10);

  /** main loop */
  do {
    if (Flag.RTC_Int != 0) {
      Flag.RTC_Int = 0;
      ES_PlaceEvent(evNewSecond);

      RTC_ReadTime(&RTC);

      /* Blink dot if show HH:MM */
      if (ES_GetState() == stShowTime) {
        RTOS_SetTask(dotOff, 500, 0);
      }
    }

    event = ES_GetEvent();
    if (event) {
      //ES_Dispatch(event);
    }

    // крутим диспетчер
    RTOS_DispatchTask();

    // делать нечего -- спим, ждём прерывание
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_mode();

  } while(1);
}

/**
 *  П о д п р о г р а м м ы
 */

/**
 * @brief Initializy perephireal
 */
static void Board_Init(void) {
  /* power off Analog Comparator */
  ACSR = ACD;

  /* GPIO */
  PORTB = BUTTON_PINS; // enable pull-up
  DDRC = DIGIT_PINS; // as output
  DDRD = (DOT_PIN | ANOD_PINS); // as output

  /* Timer2 - refresh Nixie values */
  TCCR2 = TIMER2_CS;
  TCNT2 = TIMER2_CNT;
  TIMSK = _BV(TOIE2);

  /* Interrupt from RTC */
  MCUCR = _BV(ISC11); // falling edge
  GICR = _BV(INT1);

  /* Enable Interrupts */
  sei();
}

static void dotOn(void) {
  PORTD |= DOT_PIN;
}

static void dotOff(void) {
  //PORTD &= ~(DOT_PIN);
}

void dotOnPersistent(void) {
  RTOS_DeleteTask(dotOff);
  PORTD |= DOT_PIN;
}

/**
  * @brief  Обработка кнопок.
  * @param  : None
  * @retval : None
  */
static void btnProcess(void) {
  uint8_t i;
  for (i=0; i<BTN_NUM; i++) {
    if (Button[i].pin != 0) {
      // button pressed
      if (BUTTON_STATE(Button[i].pin) == 0) {
        Button[i].time ++;
        if (Button[i].time > BTN_TIME_HOLDED) {
          Button[i].time -= BTN_TIME_REPEATED;
          ES_PlaceEvent(Button[i].holded); // button pressed auto repeat
        }
      } else {
        // button released
        if (Button[i].time >= BTN_TIME_PRESSED) {
          ES_PlaceEvent(Button[i].pressed); // process short press.
        }
        Button[i].time = 0;
      }
    } /* end (pin == 0) */
  } /* end FOR */
}

void showTime(void) {
  dotOn();
  Digit[0] = RTC.Hr >> 4;
  Digit[1] = RTC.Hr & 0x0F;
  Digit[2] = RTC.Min >> 4;
  Digit[3] = RTC.Min & 0x0F;
}

void showMMSS(void) {
  Digit[0] = RTC.Min >> 4;
  Digit[1] = RTC.Min & 0x0F;
  Digit[2] = RTC.Sec >> 4;
  Digit[3] = RTC.Sec & 0x0F;
}

void showWDay(void) {
  PORTD &= ~(DOT_PIN);
  Digit[0] = DIGIT_BLANK;
  Digit[1] = DIGIT_BLANK;
  Digit[2] = RTC.WD & 0x0F;
  Digit[3] = DIGIT_BLANK;
}

void showMDay(void) {
  PORTD &= ~(DOT_PIN);
  Digit[0] = RTC.Day >> 4;
  Digit[1] = RTC.Day & 0x0F;
  Digit[2] = DIGIT_BLANK;
  Digit[3] = DIGIT_BLANK;
}

void showMonth(void) {
  PORTD &= ~(DOT_PIN);
  Digit[0] = DIGIT_BLANK;
  Digit[1] = DIGIT_BLANK;
  Digit[2] = RTC.Mon >> 4;
  Digit[3] = RTC.Mon & 0x0F;
}

void showYear(void) {
  PORTD &= ~(DOT_PIN);
  Digit[0] = 0x02;
  Digit[1] = 0x00;
  Digit[2] = RTC.Year >> 4;
  Digit[3] = RTC.Year & 0x0F;
}


/**
 *  П р е р ы в а н и я
 */

/**
 * @brief RTC one seconds interrupt
 */
ISR (INT1_vect) {
  Flag.RTC_Int = 1;
}

/**
 * @brief Refresh Nixie output
 * @note Digit[] must be in range 0x00 - 0x0F
 */
ISR(TIMER2_OVF_vect) {
  static uint8_t idx = 0;

  // reload timer
  TCNT2 = TIMER2_CNT;

  // read current register value and clean bits
  uint8_t pd = PORTD & ~ANOD_PINS;
  uint8_t pc = PORTC & ~DIGIT_PINS;

  // power off lamps
  PORTD = pd;
  PORTC = pc;

  switch (idx) {
  case 0:
    // output lamp value
    PORTC = pc | Digit[0];
    // power on lamp
    PORTD = pd | ANOD1;
    idx = 1;
    break;

  case 1:
    PORTC = pc | Digit[1];
    PORTD = pd | ANOD2;
    idx = 2;
    break;

  case 2:
    PORTC = pc | Digit[2];
    PORTD = pd | ANOD3;
    idx = 3;
    break;

  case 3:
    PORTC = pc | Digit[3];
    PORTD = pd | ANOD4;
    idx = 0;
    break;

  default:
    idx = 0;
    break;
  }
}

/**
 * @brief заглушка для неиспользуемых прерываний
 */
ISR(__vector_default,ISR_NAKED) {
  reti();
}
